#include "treebasics.h"

struct TreeSet 
{
    int size;
    int value_type;  // 0: int, 1: bool, 2: float, 3: string
    struct TreeNode* root;
};

struct TreeSet* create_empty_set(int type_id)
{
    struct TreeSet* treeset = (struct TreeSet*)malloc(sizeof(struct TreeSet));
    treeset->size = 0;
    treeset->value_type = type_id;
    treeset->root = NULL;
    return treeset;
}

struct TreeNode* create_node(struct TreeNode* parent, union Value value)
{
    struct TreeNode* node = (struct TreeNode*)malloc(sizeof(struct TreeNode));
    node->value = value;
    node->parent = parent;
    node->lc = NULL;
    node->rc = NULL;
    return node;
}

void add_elem(struct TreeSet* treeset, void* value_ptr) 
{
    union Value uvalue = parse_value(value_ptr, treeset->value_type);
    if (treeset->root == NULL)
    {
        treeset->root = create_node(NULL, uvalue);
        treeset->size = 1;
        return;
    }
    struct TreeNode* curr_node = treeset->root;
    while (true)
    {
        if (uvalue_equal(uvalue, curr_node->value, treeset->value_type)) return;
        if (uvalue_lt(uvalue, curr_node->value, treeset->value_type))
        {
            if (curr_node->lc == NULL)
            {
                curr_node->lc = create_node(curr_node, uvalue);
                treeset->size++;
                return;
            }
            else
            {
                curr_node = curr_node->lc;
            }
        }
        else
        {
            if (curr_node->rc == NULL)
            {
                curr_node->rc = create_node(curr_node, uvalue);
                treeset->size++;
                return;
            }
            else
            {
                curr_node = curr_node->rc;
            }
        }
    }
}

struct TreeNode* search_elem_internal(struct TreeSet* treeset, union Value uvalue)
{
    if (treeset->root == NULL) return NULL;
    struct TreeNode* curr_node = treeset->root;
    while (true)
    {
        if (uvalue_equal(uvalue, curr_node->value, treeset->value_type)) return curr_node;
        if (uvalue_lt(uvalue, curr_node->value, treeset->value_type))
        {
            if (curr_node->lc == NULL) return NULL;
            else
            {
                curr_node = curr_node->lc;
            }
        }
        else
        {
            if (curr_node->rc == NULL) return NULL;
            else
            {
                curr_node = curr_node->rc;
            }
        }
    }
}

bool search_elem(struct TreeSet* treeset, void* value_ptr) 
{
    union Value uvalue = parse_value(value_ptr, treeset->value_type);
    struct TreeNode* curr_node = search_elem_internal(treeset, uvalue);
    return (curr_node != NULL);
}

void delete_elem(struct TreeSet* treeset, void* value_ptr) 
{
    union Value uvalue = parse_value(value_ptr, treeset->value_type);
    // reference: https://blog.csdn.net/zxnsirius/article/details/52131433
    struct TreeNode* curr_node = search_elem_internal(treeset, uvalue);
    if (curr_node == NULL) return;
    treeset->size--;
    if (curr_node->lc == NULL)
    {
        if (curr_node == treeset->root)  // root node
        {
            treeset->root = curr_node->rc;
            if (treeset->root != NULL) treeset->root->parent = NULL;
        }
        else if (curr_node->rc == NULL)  // leaf node
        {
            if (is_left_child(curr_node)) curr_node->parent->lc = NULL;
            else curr_node->parent->rc = NULL;
        }
        else                             // middle node
        {
            if (is_left_child(curr_node)) curr_node->parent->lc = curr_node->rc;
            else curr_node->parent->rc = curr_node->rc;
            curr_node->rc->parent = curr_node->parent;
        }
    }
    else if (curr_node->rc == NULL)
    {
        if (curr_node == treeset->root)
        {
            treeset->root = curr_node->lc;
            if (treeset->root != NULL) treeset->root->parent = NULL;
        }
        else
        {
            if (is_left_child(curr_node)) curr_node->parent->lc = curr_node->lc;
            else curr_node->parent->rc = curr_node->lc;
            curr_node->lc->parent = curr_node->parent;
        }
    }
    else  // both left and right child exist, find the largest element that is smaller than the deleted element
    {
        struct TreeNode* left_neighbor = curr_node->lc;
        while (true)
        {
            if (left_neighbor->rc != NULL) left_neighbor = left_neighbor->rc;
            else break;
        }
        if (left_neighbor->lc == NULL)  // leaf node
        {
            if (is_left_child(left_neighbor)) left_neighbor->parent->lc = NULL;
            else left_neighbor->parent->rc = NULL;
        }
        else
        {
            if (is_left_child(left_neighbor)) left_neighbor->parent->lc = left_neighbor->lc;
            else left_neighbor->parent->rc = left_neighbor->lc;
            left_neighbor->lc->parent = left_neighbor->parent;
        }
        curr_node->value = left_neighbor->value;
    }
}

int get_set_size(struct TreeSet* treeset)
{
    return treeset->size;
}


// int main(void)
// {
//     struct TreeSet* set1 = create_empty_set(2);
//     float a = 3;
//     add_elem(set1, (void*)(&a));
//     a = 5;
//     add_elem(set1, (void*)(&a));
//     a = 3;
//     add_elem(set1, (void*)(&a));
//     int set1_size = get_set_size(set1);
//     printf("set1 size: %d\n", set1_size);
//     delete_elem(set1, (void*)(&a));
//     set1_size = get_set_size(set1);
//     printf("set1 size: %d\n", set1_size);
//     return 0;
// }
